/*
 * Alternative TDMoX implementation for DAHDI - sysfs code
 *
 * Written by Pavel Selivanov <pavel[AT]parabel.ru>
 *
 * Copyright (C) 2014, Parabel, Inc.
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <linux/device.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/string.h>
#include <linux/ctype.h>

#include <dahdi/kernel.h>
#include <dahdi/user.h>

#include "dahdi_tdmox.h"
#include "dahdi_tdmox-sysfs.h"

/*
	Full names (dahdi_tdmox_sysfs_) for exported functions.
	Stripped names (tdmox_sysfs_) for static(local) functions & structs.
*/

/*== SECTION: declarations ==*/

#define TDMOX_CHAN_ATTRS 10		/* Attributes per CHAN */
#define TDMOX_SPAN_ATTRS 32		/* Attributes per SPAN */

#define to_tdmox_sysfs_attr(attr_ptr) container_of(attr_ptr, struct tdmox_sysfs_attr, attr);
struct tdmox_sysfs_attr {
	struct device_attribute attr;
	char name[TDMOX_ATTR_NAME_STRLEN];
	char data[TDMOX_CHANS_STRLEN];
};

struct tdmox_sysfs_chan {
	char name[TDMOX_CHAN_NAME_STRLEN];
/*	int tdmox_span_id; Not used for now. Reserved. */
/*	int tdmox_chan_id; Not used for now. Reserved. */
	struct tdmox_sysfs_attr attr_name;
	struct tdmox_sysfs_attr attr_sig;
	struct tdmox_sysfs_attr attr_sigcap;
	struct tdmox_sysfs_attr *tsattrs[TDMOX_CHAN_ATTRS+1];	/* Pointers to attrubutes above */
	struct attribute *dev_attrs[TDMOX_CHAN_ATTRS+1];
	struct attribute_group attr_group;
	int attrs_total;
};

#define to_tdmox_sysfs_span(dev_ptr) container_of(dev_ptr, struct tdmox_sysfs_span, device);
struct tdmox_sysfs_span {
	struct device device;
	/* SPAN-specific attributes. */
	struct tdmox_sysfs_attr attr_name;
	struct tdmox_sysfs_attr attr_spantype;
	struct tdmox_sysfs_attr attr_channels;
	struct tdmox_sysfs_attr attr_driver;
	struct tdmox_sysfs_attr attr_addr;
	struct tdmox_sysfs_attr attr_desc;
	struct tdmox_sysfs_attr attr_manufacturer;
	struct tdmox_sysfs_attr attr_location;
	struct tdmox_sysfs_attr attr_devicetype;
	struct tdmox_sysfs_attr attr_flags;
	struct tdmox_sysfs_attr attr_linecompat;
	struct tdmox_sysfs_attr attr_deflaw;
	struct tdmox_sysfs_attr attr_register;
	struct tdmox_sysfs_attr attr_timing;
	struct tdmox_sysfs_attr attr_hardware_id;
	struct tdmox_sysfs_attr *attrs[TDMOX_SPAN_ATTRS+1];	/* Pointers to attrubutes above */
	int attrs_total;
	/* Channels-specific data*/
	struct tdmox_sysfs_chan chans[TDMOX_CHANS_MAX];
	struct attribute_group *dev_attr_group[TDMOX_CHANS_MAX+1];	/* tailing NULL */
	int span_id;
};

struct str_flag {
	const char *str;
	unsigned long flag;
};

/*== SECTION: global variables ==*/

static struct str_flag tdmox_law[] = {
	{"DEFAULT", DAHDI_LAW_DEFAULT},
	{"MULAW", DAHDI_LAW_MULAW},
	{"ALAW", DAHDI_LAW_ALAW},
	{NULL, DAHDI_LAW_DEFAULT}
};

static struct str_flag tdmox_flags[] = {
	{"RBS", DAHDI_FLAG_RBS},
	{NULL, 0}
};

static struct str_flag tdmox_signalling[] = {
	{"FXSLS", DAHDI_SIG_FXSLS},
	{"FXSKS", DAHDI_SIG_FXSKS},
	{"FXSGS", DAHDI_SIG_FXSGS},
	{"FXOLS", DAHDI_SIG_FXOLS},
	{"FXOKS", DAHDI_SIG_FXOKS},
	{"FXOGS", DAHDI_SIG_FXOGS},
	{"E&M", DAHDI_SIG_EM},
	{"E&M-E1", DAHDI_SIG_EM_E1},
	{"Clear", DAHDI_SIG_CLEAR},
	{"HDLCRAW", DAHDI_SIG_HDLCRAW},
	{"HDLCFCS", DAHDI_SIG_HDLCFCS},
	{"HDLCNET", DAHDI_SIG_HDLCNET},
	{"Hardware-assisted HDLC", DAHDI_SIG_HARDHDLC},
	{"MTP2", DAHDI_SIG_MTP2},
	{"Slave", DAHDI_SIG_SLAVE},
	{"CAS", DAHDI_SIG_CAS},
	{"DACS", DAHDI_SIG_DACS},
	{"DACS+RBS", DAHDI_SIG_DACS_RBS},
	{"SF (ToneOnly)", DAHDI_SIG_SF},
	{"Unconfigured", DAHDI_SIG_NONE}, /* 0 */
	{NULL, 0}
};

static struct str_flag tdmox_linecompat[] = {
	{"D4", DAHDI_CONFIG_D4},
	{"ESF", DAHDI_CONFIG_ESF},
	{"AMI", DAHDI_CONFIG_AMI},
	{"B8ZS", DAHDI_CONFIG_B8ZS},
	{"CCS", DAHDI_CONFIG_CCS},
	{"HDB3", DAHDI_CONFIG_HDB3},
	{"CRC4", DAHDI_CONFIG_CRC4},
	{"NTTE", DAHDI_CONFIG_NOTOPEN},
	{"TERM", DAHDI_CONFIG_NTTE},
	{"NOTOPEN", DAHDI_CONFIG_TERM},
	{NULL, 0}
};

static struct tdmox_sysfs_span *tdmox_spans[TDMOX_SPANS_MAX];

/*== SECTION: Code  ==*/

//TODO - correct ?
static int tdmox_match(struct device *dev, struct device_driver *driver)
{
	return !strncmp(dev->bus->name, driver->name, strlen(driver->name));
}

static void tdmox_bus_release(struct device *dev)
{
	printk(KERN_WARNING "dahdi_tdmox bus release\n");
}

struct bus_type tdmox_bus_type = {
	.name = "dahdi_tdmox",
	.match = tdmox_match,
//TODO??	.hotplug  = tdmox_hotplug,
};

struct device tdmox_bus = {
	.bus   = &tdmox_bus_type,
	.release  = tdmox_bus_release
};

/* no references to dahdi_tdmox devices, so, fake release (for now) */
static void tdmox_dev_release(struct device *dev)
{ }

static int tdmox_strtoi(const char *s, unsigned int base, int dflt) {
	long val = 0;

	if (0 == kstrtol(s, base, &val))
		return val;
	return dflt;
}

static ssize_t show_attr(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct tdmox_sysfs_attr *tsattr = to_tdmox_sysfs_attr(attr);
	return snprintf(buf, PAGE_SIZE, "%s\n", tsattr->data);
}

static void strip_cpy(char *dst, const char *src, int len)
{
	const char *s = src;

	while(*s && isspace(*s)) s++; /* Locating first non-space symbol */
	strncpy(dst, s, len);
	dst = strstrip(dst); /* Compiler will warn us if we will not use result. */
}

static void tdmox_sysfs_clear_attrs(struct tdmox_sysfs_span *tsspan)
{
	int sa_id, ca_id, chan_id;

	for (sa_id=0; tsspan->attrs[sa_id]; ++sa_id)
		tsspan->attrs[sa_id]->data[0] = 0;

	for (chan_id=0; tsspan->dev_attr_group[chan_id]; ++chan_id) {
		struct tdmox_sysfs_chan *tschan = &(tsspan->chans[chan_id]);
		for (ca_id=0; tschan->tsattrs[ca_id]; ++ca_id) {
			tschan->tsattrs[ca_id]->data[0] = 0;
		}
	}
}

int tdmox_register(struct device *dev, struct device_attribute *attr)
{
	struct tdmox_sysfs_attr *tsattr;
	struct tdmox_sysfs_span *tsspan;
	struct dahdi_tdmox_span dds;
	int ret = 0;

	tsattr = to_tdmox_sysfs_attr(attr);
	tsspan = to_tdmox_sysfs_span(dev);

	dds.numchans = tdmox_strtoi(tsspan->attr_channels.data, 10, 0);
	strncpy(dds.driver, tsspan->attr_driver.data, sizeof(dds.driver));
	strncpy(dds.addr, tsspan->attr_addr.data, sizeof(dds.addr));
	dds.span_id_sysfs = tsspan->span_id;

	if (!strcmp(tsattr->data, "1")) {
		return dahdi_tdmox_create(&dds);
	} else {
		ret = dahdi_tdmox_destroy(&dds);
		tdmox_sysfs_clear_attrs(tsspan);
		return ret;
	}
}

static ssize_t store_attr(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
{
	struct tdmox_sysfs_attr *tsattr = to_tdmox_sysfs_attr(attr);

	strip_cpy(tsattr->data, buf, sizeof(tsattr->data));

	if (!strncmp(tsattr->name, "register", sizeof(tsattr->name)))
		tdmox_register(dev, attr);

	return count;
}

static void tdmox_sysfs_init_attr(struct tdmox_sysfs_attr *attr, const char *name)
{
	if (!attr)
		return;

	strncpy(attr->name, name, sizeof(attr->name));
	attr->attr.attr.name = attr->name;
//	attr->owner = driver->module;
	attr->attr.attr.mode = S_IRUSR | S_IWUSR;
	attr->attr.show = show_attr;
	attr->attr.store = store_attr;	
}

static int tdmox_sysfs_create_attrs(struct device *dev, struct tdmox_sysfs_span *tsspan)
{
	int ret = 0, attr_id;

	for (attr_id=0; tsspan->attrs[attr_id]; ++attr_id) {
		ret = device_create_file(dev, &(tsspan->attrs[attr_id]->attr));
		if (ret) {
			/* TODO: Check this (simulate), print error here */
			while(attr_id--)
				device_remove_file(dev, &(tsspan->attrs[attr_id]->attr));
			return ret;
		}
	}
	return ret;
}

static void tdmox_sysfs_delete_attrs(struct device *dev, struct tdmox_sysfs_span *tsspan)
{
	int attr_id;

	for (attr_id=0; tsspan->attrs[attr_id]; ++attr_id) {
		device_remove_file(dev, &(tsspan->attrs[attr_id]->attr));
	}
}

static void tdmox_sysfs_addattr(struct tdmox_sysfs_span *tsspan, struct tdmox_sysfs_attr *attr, const char *name)
{
	if (tsspan->attrs_total >= TDMOX_SPAN_ATTRS-1) /* Tailing element MUST be NULL */
		return; //TODO: WARN here (assert ?)

	tdmox_sysfs_init_attr(attr, name);
	tsspan->attrs[tsspan->attrs_total] = attr;
	tsspan->attrs_total++;
}

static void tdmox_sysfs_chan_addattr(struct tdmox_sysfs_chan *tschan, struct tdmox_sysfs_attr *attr, const char *name)
{
	if (tschan->attrs_total >= TDMOX_CHAN_ATTRS-1)  /* Tailing element MUST be NULL */
		return; //TODO: WARN here (assert ?)

	tdmox_sysfs_init_attr(attr, name);
	tschan->tsattrs[tschan->attrs_total] = attr;
	tschan->dev_attrs[tschan->attrs_total] = &(attr->attr.attr);
	tschan->attrs_total++;
}

static int tdmox_sysfs_init_spans(void)
{
	int span_id, chan_id, ret=0;

	for (span_id=0; span_id<TDMOX_SPANS_MAX; ++span_id) {
		/* reminder: zero every array of pointers ! */
		struct tdmox_sysfs_span *tsspan = kzalloc(sizeof(struct tdmox_sysfs_span), GFP_KERNEL);
		struct device *dev = &(tsspan->device);
		
		tsspan->span_id = span_id;
		dev_set_name(dev, "span-%d", span_id+1);
		dev->bus = &tdmox_bus_type;
		dev->parent = &tdmox_bus;
	//	dev_set_drvdata(dev, span);
		dev->release = tdmox_dev_release;

		for (chan_id=0; chan_id<TDMOX_CHANS_MAX; ++chan_id) {
			struct tdmox_sysfs_chan *tschan = &(tsspan->chans[chan_id]);

			tdmox_sysfs_chan_addattr(tschan, &(tschan->attr_name), "name");
			tdmox_sysfs_chan_addattr(tschan, &(tschan->attr_sig), "sig");
			tdmox_sysfs_chan_addattr(tschan, &(tschan->attr_sigcap), "sigcap");

			snprintf (tschan->name, sizeof(tschan->name), "chan-%d", chan_id+1);
			tschan->attr_group.name = tschan->name;
			tschan->attr_group.attrs = tschan->dev_attrs;
//TODO: I don't like to have casting in kernel, as I can miss sirious bug !
			tsspan->dev_attr_group[chan_id] = &(tschan->attr_group);
		}
		tsspan->device.groups = tsspan->dev_attr_group;

		ret = device_register(dev);
		if (ret) {
			printk(KERN_WARNING "device register failed\n");
			while(span_id>0)
				device_unregister(&(tdmox_spans[span_id-1]->device));
			return ret;
		}
		tdmox_spans[span_id] = tsspan;

		tdmox_sysfs_addattr(tsspan, &(tsspan->attr_addr), 	"addr");
		tdmox_sysfs_addattr(tsspan, &(tsspan->attr_channels), "channels");
		tdmox_sysfs_addattr(tsspan, &(tsspan->attr_deflaw), 	"deflaw");
		tdmox_sysfs_addattr(tsspan, &(tsspan->attr_desc), 	"desc");
		tdmox_sysfs_addattr(tsspan, &(tsspan->attr_devicetype), "devicetype");
		tdmox_sysfs_addattr(tsspan, &(tsspan->attr_driver), 	"driver");
		tdmox_sysfs_addattr(tsspan, &(tsspan->attr_flags), 	"flags");
		tdmox_sysfs_addattr(tsspan, &(tsspan->attr_linecompat), "linecompat");
		tdmox_sysfs_addattr(tsspan, &(tsspan->attr_location), "location");
		tdmox_sysfs_addattr(tsspan, &(tsspan->attr_manufacturer), "manufacturer");
		tdmox_sysfs_addattr(tsspan, &(tsspan->attr_name), 	"name");
		tdmox_sysfs_addattr(tsspan, &(tsspan->attr_spantype), "spantype");
		tdmox_sysfs_addattr(tsspan, &(tsspan->attr_timing), "timing");
		tdmox_sysfs_addattr(tsspan, &(tsspan->attr_register), "register");

		ret = tdmox_sysfs_create_attrs(dev, tsspan);
		if (ret)
			return ret;
	}
	return ret;
}

static void tdmox_copy_tsattr(char *dst, const struct tdmox_sysfs_attr *tsattr, int len)
{
	if (strlen(tsattr->data) > 0)
		strlcpy(dst, tsattr->data, len);
}

static void tdmox_ptr_tsattr(const char **dst, const struct tdmox_sysfs_attr *tsattr)
{
	if (strlen(tsattr->data) > 0)
		*dst = tsattr->data;
}

static void tdmox_parse_spantype (const char *str, enum spantypes *type)
{
	char data[TDMOX_CHANS_STRLEN];

	if (!strlen(str))
		return;

	*type = SPANTYPE_INVALID;
	if (sscanf(str, "%s", data) > 0)
		*type = dahdi_str2spantype(data);
}

/* up to unsigned long bits in flags */
void tdmox_parse_flags(const char *str, const struct str_flag *str2flag, void *res_flag)
{
	char data[TDMOX_CHANS_STRLEN], *stringp, *pdata;
	int i;

	if (!strlen(str))
		return;
	*((unsigned long *)res_flag) = 0;
	strncpy(data, str, TDMOX_CHANS_STRLEN);


	stringp = data;
	while ( (pdata = strsep(&stringp, " \r\t\n")) )  {
		for (i=0; NULL != str2flag[i].str; ++i) {
			if (!strcmp(pdata, str2flag[i].str)) {
				*((unsigned long *)res_flag) |= str2flag[i].flag;
			}
		}
	}
}

void tdmox_dump_dynamic(const struct dahdi_tdmox *d)
{
	const struct dahdi_span *span = &(d->span);

	printk(KERN_WARNING "Dumping dahdi_tdmox...\n");
	printk(KERN_WARNING "d::addr: %s\n", d->addr);
	printk(KERN_WARNING "d::dname: %s\n", d->dname);
	printk(KERN_WARNING "ddev::devicetype: %s\n", d->ddev->devicetype);
	printk(KERN_WARNING "ddev::hardware_id: %s\n", d->ddev->hardware_id);
	printk(KERN_WARNING "ddev::location: %s\n", d->ddev->location);
	printk(KERN_WARNING "ddev::manufacturer: %s\n", d->ddev->manufacturer);

	printk(KERN_WARNING "span::deflaw: %x\n", span->deflaw);
	printk(KERN_WARNING "span::desc: %s\n", span->desc);
	printk(KERN_WARNING "span::flags: %lx\n", span->flags);
	printk(KERN_WARNING "span::linecompat: %x\n", span->linecompat);
	printk(KERN_WARNING "span::name: %s\n", span->name);
	printk(KERN_WARNING "span::spantype: %x\n", span->spantype);
}

int dahdi_tdmox_sysfs_cfg_ddev(struct dahdi_tdmox *d, int span_id_sysfs)
{
	int ret=0, sigcap=0, chan_id=0;
	struct dahdi_span *span = &(d->span);
	struct tdmox_sysfs_span *tsspan = tdmox_spans[span_id_sysfs];
	const char *name_prefix=NULL;

/*	driver (dname) and addr have to be filled by create_span, but, ...
 *  Doing once again, in case, if someone will remove it from create_dynamic. */
	tdmox_copy_tsattr(d->addr, &(tsspan->attr_addr), sizeof(d->addr));
	tdmox_copy_tsattr(d->dname, &(tsspan->attr_driver), sizeof(d->dname));

/* 	TODO: Think about defaults a lot. 
	Digital or Analog (linecompat) ?
	chan->sig ?
	chan->sigcap ?
	NT/TE ?
*/
/*	d->span.linecompat = DAHDI_CONFIG_HDB3 | DAHDI_CONFIG_CCS | DAHDI_CONFIG_CRC4 | DAHDI_CONFIG_AMI; */
	tdmox_parse_flags(tsspan->attr_linecompat.data, tdmox_linecompat, &(span->linecompat));	//int
	d->span.deflaw = DAHDI_LAW_ALAW;
	sigcap = DAHDI_SIG_EM |
			      DAHDI_SIG_FXSLS | DAHDI_SIG_FXSKS |
			      DAHDI_SIG_FXSGS | DAHDI_SIG_FXOLS |
			      DAHDI_SIG_FXOKS | DAHDI_SIG_FXOGS |
			      DAHDI_SIG_SF | DAHDI_SIG_DACS_RBS |
			      DAHDI_SIG_CAS | DAHDI_SIG_CLEAR;

	if (d->span.linecompat) {
    	d->span.spantype = SPANTYPE_DIGITAL_E1;
    	/* Should I add RBS flag by default ? Possible to disable via sysfs ? */
		sigcap |= DAHDI_SIG_EM_E1;
		name_prefix = "TDM_D";
	} else {
    	d->span.spantype = SPANTYPE_ANALOG_MIXED;
		d->span.flags |= DAHDI_FLAG_RBS;
		name_prefix = "TDM_A";
	}
	snprintf(span->name, sizeof(span->name), "%s/%s/%s",
		 name_prefix, d->dname, d->addr);
	snprintf(span->desc, sizeof(span->desc),
		 "Dynamic '%s' span at '%s'", d->dname, d->addr);

	d->ddev->devicetype = d->span.name;
	d->ddev->hardware_id = d->span.name;

	for (chan_id = 0; chan_id < d->span.channels; chan_id++) {
		struct tdmox_sysfs_chan *tschan = &(tsspan->chans[chan_id]);
		struct dahdi_chan *dchan = d->chans[chan_id];
		const char *linetype="";

		dchan->sigcap = sigcap;
		tdmox_copy_tsattr(dchan->name, &(tschan->attr_name), sizeof(dchan->name));
		tdmox_parse_flags(tschan->attr_sigcap.data, tdmox_signalling, &(dchan->sigcap));
		tdmox_parse_flags(tschan->attr_sig.data, tdmox_signalling, &(dchan->sig));
		if (dchan->sig) {
			/* Seems user know, what he wants.
			dahdi_genconf(Chans.pm) will probe "$type = 'FXO' if $signalling =~ /^FXS/;"
			But, dahdi_cfg will see no diff for the channel, and will not ioctl(DAHDI_CHANCONFIG).
			 Let's do it manually. Any ideas ?
			 channel name= TDM_A_S(FXS), TDM_A_O(FXO), TDM_A_E(E&M) and patch Chans.pm ?*/
			if (dchan->sig & DAHDI_SIG_CLEAR)
				dchan->flags |= DAHDI_FLAG_CLEAR;
			else
				dchan->flags |= DAHDI_FLAG_AUDIO;
		}
		if (dchan->sigcap & __DAHDI_SIG_FXS)
			linetype = "O";
		if (dchan->sigcap & __DAHDI_SIG_FXO)
			linetype = "S";
		if (dchan->sigcap & DAHDI_SIG_EM)
			linetype = "E";
		/* dahdi_chan->name HAVE TO BE <= 40 bytes... 
		channels > 999 or interface name > 4 bytes will overflow... */
		snprintf(dchan->name, sizeof(dchan->name), "%s%s/%s/%s/%d",
			name_prefix, linetype, d->dname, d->addr, chan_id+1);

		dchan->chanpos = chan_id + 1;
		dchan->pvt = d;
	}

	d->timing = tdmox_strtoi(tsspan->attr_timing.data, 10, 0);

	tdmox_copy_tsattr(span->desc, &(tsspan->attr_desc), sizeof(span->desc));
	tdmox_copy_tsattr(span->name, &(tsspan->attr_name), sizeof(span->name));

	tdmox_ptr_tsattr(&(d->ddev->devicetype), &(tsspan->attr_devicetype));
	tdmox_ptr_tsattr(&(d->ddev->hardware_id), &(tsspan->attr_hardware_id));
	tdmox_ptr_tsattr(&(d->ddev->location), &(tsspan->attr_location));
	tdmox_ptr_tsattr(&(d->ddev->manufacturer), &(tsspan->attr_manufacturer));

	tdmox_parse_spantype(tsspan->attr_spantype.data, &(span->spantype));
	tdmox_parse_flags(tsspan->attr_flags.data, tdmox_flags, &(span->flags));	//unsigned long
	/* We have to parse linecompat at head. Moved there */
	tdmox_parse_flags(tsspan->attr_deflaw.data, tdmox_law, &(span->deflaw));	//int

	tdmox_dump_dynamic(d);

	return ret;
}

int tdmox_sysfs_init(void)
{
	int ret;

	memset(tdmox_spans, 0, sizeof(tdmox_spans));

	ret = bus_register(&tdmox_bus_type);
	if (ret) {
		printk(KERN_ERR "Failed to register dahdi_tdmox bus\n");
		return ret;
	}
	dev_set_name(&tdmox_bus, "dahdi_tdmox");
	ret = device_register(&tdmox_bus);
	if (ret) {
		printk(KERN_NOTICE "Unable to register dahdi_tdmox bus device\n");
		bus_unregister(&tdmox_bus_type);
		return ret;
	}
	ret = tdmox_sysfs_init_spans();
	if (ret) {
		printk(KERN_NOTICE "Unable to register SPANs in dahdi_tdmox\n");
		device_unregister(&tdmox_bus);
		bus_unregister(&tdmox_bus_type);
		return ret;
	}
	printk(KERN_INFO "dahdi_tdmox-sysfs initialized\n");
	return ret;
}

void tdmox_sysfs_cleanup(void)
{
	int i;

	for (i=0; i<TDMOX_SPANS_MAX; ++i) {
		struct tdmox_sysfs_span *sspan = tdmox_spans[i];
		if (sspan) {
			struct device *dev = &(sspan->device);
			tdmox_sysfs_delete_attrs(dev, sspan);
			device_unregister(dev);
			kfree(sspan);
			tdmox_spans[i] = NULL;
		}
	}
	device_unregister(&tdmox_bus);
	bus_unregister(&tdmox_bus_type);
	printk(KERN_INFO "dahdi_tdmox-sysfs cleanup\n");
}

EXPORT_SYMBOL(dahdi_tdmox_sysfs_cfg_ddev);
EXPORT_SYMBOL(dahdi_tdmox_sysfs_init);
EXPORT_SYMBOL(dahdi_tdmox_sysfs_cleanup);
